# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simple unit test for Pipeline visualization using Mermaid."""

import io
import unittest
from unittest.mock import MagicMock, patch

import lsst.pipe.base.connectionTypes as cT
import lsst.pipe.base.pipeline_graph.visualization as vis
import lsst.utils.tests
from lsst.pipe.base import Pipeline, PipelineTask, PipelineTaskConfig, PipelineTaskConnections
from lsst.pipe.base.mermaid_tools import pipeline2mermaid

MERMAID_AVAILABLE = vis._mermaid.MERMAID_AVAILABLE

if MERMAID_AVAILABLE:
    # Since we are patching Mermaid’s `__init__`, calling
    # `vis.Mermaid.__init__` inside the mock would cause infinite recursion.
    # Using a reference to the original initializer prevents this.
    _originalMermaidInit = vis._mermaid.Mermaid.__init__

    # Mocked content for SVG and PNG files.
    MOCKED_SVG_CONTENT = b"<svg>Mocked SVG Content</svg>"
    MOCKED_PNG_CONTENT = b"\x89PNG\r\n\x1a\nMocked PNG Content"

    def _mockMermaidInit(self, *args, **kwargs):
        # Call the original initializer to set things up.
        _originalMermaidInit(self, *args, **kwargs)

        # Override only the `svg_response` and `img_response` attributes.
        self.svg_response = MagicMock(content=MOCKED_SVG_CONTENT)
        self.img_response = MagicMock(content=MOCKED_PNG_CONTENT)


class ExamplePipelineTaskConnections(PipelineTaskConnections, dimensions=("visit", "detector")):
    """Connections class used for testing.

    Parameters
    ----------
    config : `~lsst.pipe.base.PipelineTaskConfig`
        The config to use for this connections class.
    """

    input1 = cT.Input(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExposureF",
        doc="Input for this task",
    )
    input1Prerequisite = cT.PrerequisiteInput(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExposureF",
        doc="Prerequisite input for this task",
    )
    input2 = cT.Input(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExposureF",
        doc="Input for this task",
    )
    input2Prerequisite = cT.PrerequisiteInput(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExposureF",
        doc="Prerequisite input for this task",
    )
    output1 = cT.Output(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExposureF",
        doc="Output for this task",
    )
    output2 = cT.Output(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExposureF",
        doc="Output for this task",
    )

    def __init__(self, *, config=None):
        super().__init__(config=config)
        for x in config.connections:
            # Avoid complaints about incorrect storage class for metadata.
            if getattr(config.connections, x).endswith("_metadata"):
                getattr(self, x).__dict__["storageClass"] = "TaskMetadata"

            # Remove inputs and outputs that are not used.
            if x.startswith("input"):
                if not getattr(config.connections, x):
                    if x in self.inputs:
                        self.inputs.remove(x)
                    elif x in self.prerequisiteInputs:
                        self.prerequisiteInputs.remove(x)
            elif x.startswith("output"):
                if not getattr(config.connections, x):
                    if x in self.outputs:
                        self.outputs.remove(x)


class ExamplePipelineTaskConfig(PipelineTaskConfig, pipelineConnections=ExamplePipelineTaskConnections):
    """Example config used for testing."""


def _makeConfig(inputNames, outputNames, pipeline, label, inputTypes=None):
    """Configure pipeline connections by adding config overrides.

    Parameters
    ----------
    inputNames : `list` or `tuple` of `str`
        A list or tuple containing up to two input names.
    outputNames : `list` or `tuple` of `str`
        A list or tuple containing up to two output names.
    pipeline : `~lsst.pipe.base.Pipeline`
        The pipeline object where configuration overrides will be applied.
    label : `str`
        The label associated with the configuration.
    inputTypes : `list` or `tuple` of `str`, optional
        A list or tuple specifying input types. Elements can be 'Prerequisite'
        or empty strings (the default, indicating a regular input).
    """
    # Ensure at least two elements, default to empty strings.
    inputTypes = (inputTypes or ["", ""])[:2]

    for i in (0, 1):
        inputName = inputNames[i] if i < len(inputNames) else ""
        isPrerequisite = inputTypes[i] == "Prerequisite" if i < len(inputNames) else False

        pipeline.addConfigOverride(label, f"connections.input{i + 1}", "" if isPrerequisite else inputName)
        pipeline.addConfigOverride(
            label, f"connections.input{i + 1}Prerequisite", inputName if isPrerequisite else ""
        )

        outputName = outputNames[i] if i < len(outputNames) else ""
        pipeline.addConfigOverride(label, f"connections.output{i + 1}", outputName)


class ExamplePipelineTask(PipelineTask):
    """Example pipeline task used for testing."""

    ConfigClass = ExamplePipelineTaskConfig
    _DefaultName = "examplePipelineTask"


def _makePipeline(tasks, inputTypeOverrides=None):
    """Generate a Pipeline instance and return it along with task definitions.

    Parameters
    ----------
    tasks : `list` of `tuple`
        Each tuple in the list has 3 or 4 items:
        - input DatasetType name(s), string or tuple of strings
        - output DatasetType name(s), string or tuple of strings
        - task label, string
        - optional task class object, can be None
    inputTypeOverrides : `dict`, optional
        Dictionary of input type overrides.

    Returns
    -------
    pipe: `~lsst.pipe.base.Pipeline`
        The pipeline instance.
    task_defs: `list` of `~lsst.pipe.base.TaskDef`
        List of task definitions.
    """
    pipe = Pipeline("test pipeline")
    for task in tasks:
        inputs = task[0]
        outputs = task[1]
        label = task[2]
        klass = task[3] if len(task) > 3 else ExamplePipelineTask
        pipe.addTask(klass, label)
        if not isinstance(inputs, tuple | list):
            inputs = (inputs,)
        if not isinstance(outputs, tuple | list):
            outputs = (outputs,)
        if inputTypeOverrides:
            inputTypes = [inputTypeOverrides.get(x) for x in inputs]
        _makeConfig(inputs, outputs, pipe, label, inputTypes)
    taskDefs = list(pipe.to_graph()._iter_task_defs())
    return pipe, taskDefs


class MermaidTestCase(unittest.TestCase):
    """A test case for Mermaid pipeline visualization."""

    def setUp(self):
        # Create a pipeline with example tasks and dataset types.
        self.pipeline, self.pipelineTaskDefs = _makePipeline(
            [
                ("A", ("B", "C"), "task0"),
                ("B", "E", "task1"),
                ("B", ("D", "F"), "task2"),
                ("D", "G", "task3"),
                (("C", "F"), "H", "task4"),
                ("task4_metadata", "I", "task5"),
            ],
            inputTypeOverrides={"A": "Prerequisite"},
        )

        # Create a pipeline graph for processing.
        self.graph = self.pipeline.to_graph(visualization_only=True)

    def validateMermaidSource(self, file):
        """Validate the Mermaid source output for basic components.

        Parameters
        ----------
        file : `~io.StringIO`
            The in-memory file-like object containing the Mermaid source.
        """
        # It's hard to validate the complete output, just checking some basic
        # things, but even that is not terribly stable.
        fileValue = file.getvalue()
        lines = fileValue.strip().split("\n")
        nClassDefs = 2
        nTasks = 6
        nTaskClasses = 6
        nDatasets = 10
        nDatasetClasses = 10
        nEdges = 16
        nLinkStyles = 2  # For the default and pre-requisite edges
        nExtra = 1  # 'flowchart' opening line

        self.assertEqual(
            len(lines),
            nClassDefs + nTasks + nTaskClasses + nDatasets + nDatasetClasses + nEdges + nLinkStyles + nExtra,
        )

        # Confirm Mermaid syntax begins with a top-down flowchart declaration.
        self.assertTrue(fileValue.startswith("flowchart TD"))

        # Select the reference set based on naming style.
        namingRef = {
            "edge1": "B:0 --> task1:2",
            "edge2": "task2:2 --> F:0",
            "classDef": "class E:0 dsType;",
            "metadata": "task4:2 --> task4_metadata:0",
        }

        # Make sure components are connected appropriately.
        self.assertIn(namingRef["edge1"], fileValue)
        self.assertIn(namingRef["edge2"], fileValue)

        # Make sure class definitions are created for datasets.
        self.assertIn(namingRef["classDef"], fileValue)

        # Make sure there is a connection created for metadata.
        self.assertIn(namingRef["metadata"], fileValue)

    def test_pipeline2mermaid(self):
        """Validate Mermaid syntax generated by `pipeline2mermaid`."""
        for source in [self.pipeline, self.pipelineTaskDefs]:
            file = io.StringIO()
            pipeline2mermaid(source, file, expand_dimensions=False)
            self.validateMermaidSource(file)

    def test_show_mermaid_source(self):
        """Test generating Mermaid source using `show_mermaid`."""
        # Validate Mermaid source output (mmd format) in a text stream.
        file = io.StringIO()
        vis.show_mermaid(self.graph, file, dataset_types=True, task_classes="full")
        self.validateMermaidSource(file)

    @unittest.skipIf(not MERMAID_AVAILABLE, "Skipping image rendering tests since `mermaid-py` is missing.")
    def test_show_mermaid_image(self):
        """Test the output of the `show_mermaid` method in png and svg formats.

        For generating image outputs, the `show_mermaid` method invokes a
        remote rendering service (`mermaid.ink`), which can lead to timeouts at
        times. To bypass this instability, we patch Mermaid's `__init__` method
        to override its `svg_response` and `img_response` attributes, allowing
        us to return predictable, mocked contents instead.
        """
        # Test image rendering formats with mocked responses in binary streams.
        with patch.object(vis._mermaid.Mermaid, "__init__", new=_mockMermaidInit):
            for fmt, expected in [("svg", MOCKED_SVG_CONTENT), ("png", MOCKED_PNG_CONTENT)]:
                file = io.BytesIO()
                vis.show_mermaid(self.graph, file, output_format=fmt, dataset_types=True, task_classes="full")
                file.seek(0)  # Read from the beginning.
                self.assertEqual(file.read(), expected)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Generic file handle leak check."""


def setup_module(module):
    """Set up the module for pytest.

    Parameters
    ----------
    module : `~types.ModuleType`
        Module to set up.
    """
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility to facilitate testing of pipelines consisting of multiple steps."""

__all__ = ["PipelineStepTester"]

import dataclasses
import unittest
import warnings

from lsst.daf.butler import Butler, DatasetType
from lsst.pipe.base import Pipeline, PipelineGraph
from lsst.resources import ResourcePath, ResourcePathExpression


@dataclasses.dataclass
class PipelineStepTester:
    """Utility class which facilitates testing of pipelines, optionally
    consisting of multiple steps.

    Two sets will be constructed by looping over the entire pipeline or all
    named subsets within the pipeline: `pure_inputs` and `all_outputs`.

    The `pure_inputs` set consists of all inputs which must be constructed and
    provided as an input into the pipeline (i.e., they will not be generated
    by the named pipeline).

    The `all_outputs` set consists of all dataset types which are generated by
    the named pipeline, either as intermediates or as final outputs.

    These sets will be checked against user-supplied sets to ensure that the
    named pipeline may still be run without raising a missing data error.

    Attributes
    ----------
    filename : `lsst.resources.ResourcePathExpression`
        The full path or URI to the pipeline YAML.
    step_suffixes : `list` [`str`]
        A list, in the order of data reduction, of the step subsets to check.
        Must include any initial "#".
    initial_dataset_types : `list` [`tuple`]
        Dataset types which require initial registry by the butler. Each
        element must be a tuple of the type name, dimensions, storage class,
        and calibration flag, as described in the constructor for
        `~lsst.daf.butler.DatasetType`. All tuple elements are required.
    expected_inputs : `set` [`str`]
        Dataset types expected as an input into the pipeline.
    expected_outputs : `set` [`str`]
        Dataset types expected to be produced as an output by the pipeline.
    pipeline_patches : `dict` [`str`, `str`], optional
        Any config overrides to be applied to the pipeline before testing it.
        This is rarely appropriate, and should be reserved for fields that
        cannot have a file-level default. The key must have the form
        "task:subtask.field".
    """

    filename: ResourcePathExpression
    step_suffixes: list[str]
    initial_dataset_types: list[tuple[str, set[str], str, bool]]
    expected_inputs: set[str]
    expected_outputs: set[str]
    pipeline_patches: dict[str, str] = dataclasses.field(default_factory=dict)

    def register_dataset_types(self, butler: Butler) -> None:
        """Register any dataset types passed to the class constructor.

        The types registered are those specified in
        ``self.initial_dataset_types``.

        Parameters
        ----------
        butler : `lsst.daf.butler.Butler`
            The (test) butler in which to register the types.
        """
        for name, dimensions, storageClass, isCalibration in self.initial_dataset_types:
            butler.registry.registerDatasetType(
                DatasetType(
                    name,
                    dimensions,
                    storageClass=storageClass,
                    isCalibration=isCalibration,
                    universe=butler.dimensions,
                )
            )

    def run(self, butler: Butler, test_case: unittest.TestCase) -> None:
        """Run the test on all pipelines passed to the class constructor.

        Parameters
        ----------
        butler : `lsst.daf.butler.Butler`
            The butler in which to run the test. The butler must be writeable,
            and its state will be modified as part of the test.
        test_case : `unittest.TestCase`
            The test case in which this test is run.

        Raises
        ------
        AssertionError
            Raised if the pipeline requires inputs that are not in
            ``self.expected_inputs``, or fails to produce outputs that are in
            ``self.expected_outputs``.
        """
        self.register_dataset_types(butler)

        all_outputs: dict[str, DatasetType] = dict()
        pure_inputs: dict[str, str] = dict()

        file_uri = ResourcePath(self.filename, forceDirectory=False)
        for suffix in self.step_suffixes:
            step_uri = file_uri.replace(fragment=suffix.removeprefix("#"))
            step_graph = self.load_pipeline_graph(step_uri)
            step_graph.resolve(butler.registry)

            for name, _ in step_graph.iter_overall_inputs():
                if name not in all_outputs:
                    tasks = step_graph.consumers_of(name)
                    pure_inputs[name] = f"{suffix}: {', '.join(task.label for task in tasks)}"
            all_outputs.update(
                {
                    name: node.dataset_type
                    for name, node in step_graph.dataset_types.items()
                    if step_graph.producing_edge_of(name) is not None
                }
            )

            for node in step_graph.dataset_types.values():
                butler.registry.registerDatasetType(node.dataset_type)

        if not pure_inputs.keys() <= self.expected_inputs:
            missing = []
            for type_name in pure_inputs.keys() - self.expected_inputs:
                suffix = pure_inputs[type_name]
                missing.append(type_name + (f" ({suffix})" if suffix else ""))
            raise AssertionError(f"Got unexpected pure_inputs: {missing}")

        if not all_outputs.keys() >= self.expected_outputs:
            missing = list(self.expected_outputs - all_outputs.keys())
            raise AssertionError(f"Missing expected_outputs: {missing}")

    def load_pipeline_graph(self, uri: ResourcePathExpression) -> PipelineGraph:
        pipeline = Pipeline.from_uri(uri)
        for fullkey, value in self.pipeline_patches.items():
            label, key = fullkey.split(":", maxsplit=1)
            try:
                pipeline.addConfigOverride(label, key, value)
            except LookupError as e:
                warnings.warn(f"{e}, skipping configuration {fullkey}={value}", UserWarning)
        return pipeline.to_graph()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest

from lsst.pipe.base import QgraphSummary
from lsst.pipe.base.quantum_reports import ExecutionStatus, QuantumReport, Report


class ReportsTestCase(unittest.TestCase):
    """A test case for reports module."""

    def test_quantumReport(self):
        """Test for QuantumReport class."""
        dataId = {"instrument": "LSST"}
        taskLabel = "task"

        qr = QuantumReport(dataId=dataId, taskLabel=taskLabel)
        self.assertEqual(qr.status, ExecutionStatus.SUCCESS)
        self.assertEqual(qr.dataId, dataId)
        self.assertEqual(qr.taskLabel, taskLabel)
        self.assertIsNone(qr.exitCode)
        self.assertIsNone(qr.exceptionInfo)

        qr = QuantumReport(status=ExecutionStatus.TIMEOUT, dataId=dataId, taskLabel=taskLabel)
        self.assertEqual(qr.status, ExecutionStatus.TIMEOUT)

        qr = QuantumReport.from_exception(
            exception=RuntimeError("runtime error"), dataId=dataId, taskLabel=taskLabel
        )
        self.assertEqual(qr.status, ExecutionStatus.FAILURE)
        self.assertEqual(qr.dataId, dataId)
        self.assertEqual(qr.taskLabel, taskLabel)
        self.assertIsNone(qr.exitCode)
        self.assertEqual(qr.exceptionInfo.className, "RuntimeError")
        self.assertEqual(qr.exceptionInfo.message, "runtime error")

        qr = QuantumReport.from_exit_code(exitCode=0, dataId=dataId, taskLabel=taskLabel)
        self.assertEqual(qr.status, ExecutionStatus.SUCCESS)
        self.assertEqual(qr.dataId, dataId)
        self.assertEqual(qr.taskLabel, taskLabel)
        self.assertEqual(qr.exitCode, 0)
        self.assertIsNone(qr.exceptionInfo)

        qr = QuantumReport.from_exit_code(exitCode=1, dataId=dataId, taskLabel=taskLabel)
        self.assertEqual(qr.status, ExecutionStatus.FAILURE)
        self.assertEqual(qr.dataId, dataId)
        self.assertEqual(qr.taskLabel, taskLabel)
        self.assertEqual(qr.exitCode, 1)
        self.assertIsNone(qr.exceptionInfo)

    def test_report(self):
        """Test for Report class."""
        report = Report(qgraphSummary=QgraphSummary(graphID="uuid"))
        self.assertEqual(report.status, ExecutionStatus.SUCCESS)
        self.assertIsNotNone(report.cmdLine)
        self.assertIsNone(report.exitCode)
        self.assertIsNone(report.exceptionInfo)

        dataId = {"instrument": "LSST"}
        taskLabel = "task"

        qr = QuantumReport.from_exception(
            exception=RuntimeError("runtime error"), dataId=dataId, taskLabel=taskLabel
        )
        report = Report(
            status=ExecutionStatus.FAILURE, exitCode=-1, qgraphSummary=QgraphSummary(graphID="uuid")
        )
        report.set_exception(RuntimeError("runtime error"))
        report.quantaReports.append(qr)
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(report.exitCode, -1)
        self.assertEqual(report.exceptionInfo.className, "RuntimeError")
        self.assertEqual(report.exceptionInfo.message, "runtime error")
        self.assertEqual(len(report.quantaReports), 1)

    def test_json(self):
        """Test for conversion to/from JSON."""
        dataId = {"instrument": "LSST"}
        taskLabel = "task"

        qr = QuantumReport.from_exception(
            exception=RuntimeError("runtime error"), dataId=dataId, taskLabel=taskLabel
        )
        report = Report(
            status=ExecutionStatus.FAILURE, exitCode=-1, qgraphSummary=QgraphSummary(graphID="uuid")
        )
        report.set_exception(RuntimeError("runtime error"))
        report.quantaReports.append(qr)
        json = report.model_dump_json(exclude_none=True, indent=2)
        self.assertIsInstance(json, str)

        report = Report.model_validate_json(json)
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(report.exitCode, -1)
        self.assertEqual(report.exceptionInfo.className, "RuntimeError")
        self.assertEqual(report.exceptionInfo.message, "runtime error")
        self.assertEqual(len(report.quantaReports), 1)
        qr = report.quantaReports[0]
        self.assertEqual(qr.status, ExecutionStatus.FAILURE)
        self.assertEqual(qr.dataId, dataId)
        self.assertEqual(qr.taskLabel, taskLabel)
        self.assertIsNone(qr.exitCode)
        self.assertEqual(qr.exceptionInfo.className, "RuntimeError")
        self.assertEqual(qr.exceptionInfo.message, "runtime error")


if __name__ == "__main__":
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Tests for scripts/utils.py functions."""

import logging
import unittest

from lsst.pipe.base.script.utils import filter_by_existence


class FilterByExistenceTestCase(unittest.TestCase):
    """Test filter_by_existence function."""

    def setUp(self):
        # Currently function itself does not need real DatasetRefs.
        # Only the butler calls requires DatasetRefs, but butler
        # is being mocked in these tests. So using strings in place
        # of DatasetRefs.
        self.orig_refs = [f"{i:05}" for i in range(1, 11)]
        self.mock_exists = {}
        for i in range(1, 11):
            self.mock_exists[f"{i:05}"] = True if i % 3 == 0 else False
        self.filtered_refs = [f"{i:05}" for i in range(1, 11) if i % 3 != 0]

    def test_success(self):
        butler_mock = unittest.mock.Mock()
        butler_mock._datastore.knows_these.return_value = self.mock_exists
        with self.assertLogs(logging.getLogger("lsst.pipe.base"), level="VERBOSE") as cm:
            filtered = filter_by_existence(butler_mock, self.orig_refs)
            self.assertCountEqual(self.filtered_refs, filtered)
        log_messages = " ".join(cm.output)
        self.assertIn("Filtering out datasets already known to the target butler", log_messages)
        self.assertIn("number of datasets to transfer", log_messages)


if __name__ == "__main__":
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import annotations

__all__ = [
    "SeparablePipelineExecutor",
]


import datetime
import getpass
import logging
from collections.abc import Iterable
from typing import Any

import lsst.resources
from lsst.daf.butler import Butler, DatasetRef
from lsst.daf.butler._rubin.temporary_for_ingest import TemporaryForIngest

from ._quantumContext import ExecutionResources
from .all_dimensions_quantum_graph_builder import AllDimensionsQuantumGraphBuilder
from .graph import QuantumGraph
from .mp_graph_executor import MPGraphExecutor
from .pipeline import Pipeline
from .quantum_graph import PredictedQuantumGraph
from .quantum_graph_builder import QuantumGraphBuilder
from .quantum_graph_executor import QuantumGraphExecutor
from .single_quantum_executor import SingleQuantumExecutor
from .taskFactory import TaskFactory

_LOG = logging.getLogger(__name__)


class SeparablePipelineExecutor:
    """An executor that allows each step of pipeline execution to be
    run independently.

    The executor can run any or all of the following steps:

        * pre-execution initialization
        * pipeline building
        * quantum graph generation
        * quantum graph execution

    Any of these steps can also be handed off to external code without
    compromising the remaining ones.

    Parameters
    ----------
    butler : `lsst.daf.butler.Butler`
        A Butler whose ``collections`` and ``run`` attributes contain the input
        and output collections to use for processing.
    clobber_output : `bool`, optional
        If set, the pipeline execution overwrites existing output files.
        Otherwise, any conflict between existing and new outputs is an error.
    skip_existing_in : `~collections.abc.Iterable` [`str`], optional
        If not empty, the pipeline execution searches the listed collections
        for existing outputs, and skips any quanta that have run to completion
        (or have no work to do). Otherwise, all tasks are attempted (subject to
        ``clobber_output``).
    task_factory : `.TaskFactory`, optional
        A custom task factory for use in pre-execution and execution. By
        default, a new instance of `.TaskFactory` is used.
    resources : `.ExecutionResources`
        The resources available to each quantum being executed.
    raise_on_partial_outputs : `bool`, optional
        If `True` raise exceptions chained by
        `.AnnotatedPartialOutputsError` immediately, instead of
        considering the partial result a success and continuing to run
        downstream tasks.
    """

    def __init__(
        self,
        butler: Butler,
        clobber_output: bool = False,
        skip_existing_in: Iterable[str] | None = None,
        task_factory: TaskFactory | None = None,
        resources: ExecutionResources | None = None,
        raise_on_partial_outputs: bool = True,
    ):
        self._butler = Butler.from_config(
            butler=butler, collections=butler.collections.defaults, run=butler.run
        )
        if not self._butler.collections.defaults:
            raise ValueError("Butler must specify input collections for pipeline.")
        if not self._butler.run:
            raise ValueError("Butler must specify output run for pipeline.")

        self._clobber_output = clobber_output
        self._skip_existing_in = list(skip_existing_in) if skip_existing_in else []

        self._task_factory = task_factory if task_factory else TaskFactory()
        self.resources = resources
        self.raise_on_partial_outputs = raise_on_partial_outputs

    def pre_execute_qgraph(
        self,
        graph: QuantumGraph | PredictedQuantumGraph,
        register_dataset_types: bool = False,
        save_init_outputs: bool = True,
        save_versions: bool = True,
    ) -> None:
        """Run pre-execution initialization.

        This method will be deprecated after DM-38041, to be replaced with a
        method that takes either a `.Pipeline` or a
        resolved `.pipeline_graph.PipelineGraph` instead of a `.QuantumGraph`.

        Parameters
        ----------
        graph : `.QuantumGraph` or `.quantum_graph.PredictedQuantumGraph`
            The quantum graph defining the pipeline and datasets to
            be initialized.
        register_dataset_types : `bool`, optional
            If `True`, register all output dataset types from the pipeline
            represented by ``graph``.
        save_init_outputs : `bool`, optional
            If `True`, create init-output datasets in this object's output run.
        save_versions : `bool`, optional
            If `True`, save a package versions dataset.
        """
        if register_dataset_types:
            graph.pipeline_graph.register_dataset_types(self._butler, include_packages=save_versions)
        if save_init_outputs:
            graph.write_init_outputs(self._butler, skip_existing=(self._butler.run in self._skip_existing_in))
            graph.write_configs(self._butler)
        if save_versions:
            graph.write_packages(self._butler)

    def make_pipeline(self, pipeline_uri: str | lsst.resources.ResourcePath) -> Pipeline:
        """Build a pipeline from pipeline and configuration information.

        Parameters
        ----------
        pipeline_uri : `str` or `lsst.resources.ResourcePath`
            URI to a file containing a pipeline definition. A URI fragment may
            be used to specify a subset of the pipeline, as described in
            :ref:`pipeline-running-intro`.

        Returns
        -------
        pipeline : `.Pipeline`
            The fully-built pipeline.
        """
        return Pipeline.from_uri(pipeline_uri)

    def make_quantum_graph_builder(
        self,
        pipeline: Pipeline,
        where: str = "",
        *,
        builder_class: type[QuantumGraphBuilder] = AllDimensionsQuantumGraphBuilder,
        **kwargs: Any,
    ) -> QuantumGraphBuilder:
        """Initialize a quantum graph builder from a pipeline and input
        datasets.

        Parameters
        ----------
        pipeline : `.Pipeline`
            The pipeline for which to generate a quantum graph.
        where : `str`, optional
            A data ID query that constrains the quanta generated.  Must not be
            provided if a custom ``builder_class`` is given and that class does
            not accept ``where`` as a construction argument.
        builder_class : `type` [ \
                `.quantum_graph_builder.QuantumGraphBuilder` ], optional
            Quantum graph builder implementation.  Ignored if ``builder`` is
            provided.
        **kwargs
            Additional keyword arguments are forwarded to ``builder_class``
            when a quantum graph builder instance is constructed.  All
            arguments accepted by the
            `~.quantum_graph_builder.QuantumGraphBuilder` base
            class are provided automatically (from explicit arguments to this
            method and executor attributes) and do not need to be included
            as keyword arguments.

        Returns
        -------
        builder : `.quantum_graph_builder.QuantumGraphBuilder`
            A quantum graph builder.
        """
        if where:
            # Only pass 'where' if it's actually provided, since some
            # QuantumGraphBuilder subclasses may not accept it.
            kwargs["where"] = where
        return builder_class(
            pipeline.to_graph(),
            self._butler,
            skip_existing_in=self._skip_existing_in,
            clobber=self._clobber_output,
            **kwargs,
        )

    def make_quantum_graph(
        self,
        pipeline: Pipeline,
        where: str = "",
        *,
        builder_class: type[QuantumGraphBuilder] = AllDimensionsQuantumGraphBuilder,
        attach_datastore_records: bool = False,
        **kwargs: Any,
    ) -> QuantumGraph:
        """Build a quantum graph from a pipeline and input datasets.

        This returns an instance of the old `.QuantumGraph` class.  Use
        `build_quantum_graph` to construct a
        `.quantum_graph.PredictedQuantumGraph`.

        Parameters
        ----------
        pipeline : `.Pipeline`
            The pipeline for which to generate a quantum graph.
        where : `str`, optional
            A data ID query that constrains the quanta generated.  Must not be
            provided if a custom ``builder_class`` is given and that class does
            not accept ``where`` as a construction argument.
        builder_class : `type` [ \
                `.quantum_graph_builder.QuantumGraphBuilder` ], optional
            Quantum graph builder implementation.  Ignored if ``builder`` is
            provided.
        attach_datastore_records : `bool`, optional
            Whether to attach datastore records.  These are currently used only
            by `lsst.daf.butler.QuantumBackedButler`, which is not used by
            `SeparablePipelineExecutor` for execution.
        **kwargs
            Additional keyword arguments are forwarded to ``builder_class``
            when a quantum graph builder instance is constructed.  All
            arguments accepted by the
            `~.quantum_graph_builder.QuantumGraphBuilder` base
            class are provided automatically (from explicit arguments to this
            method and executor attributes) and do not need to be included
            as keyword arguments.

        Returns
        -------
        graph : `.QuantumGraph`
            The quantum graph for ``.Pipeline`` as run on the datasets
            identified by ``where``.

        Notes
        -----
        This method does no special handling of empty quantum graphs. If
        needed, clients can use `len` to test if the returned graph is empty.
        """
        metadata = {
            "input": self._butler.collections.defaults,
            "output_run": self._butler.run,
            "skip_existing_in": self._skip_existing_in,
            "skip_existing": bool(self._skip_existing_in),
            "data_query": where,
            "user": getpass.getuser(),
            "time": str(datetime.datetime.now()),
        }
        qg_builder = self.make_quantum_graph_builder(pipeline, where, builder_class=builder_class, **kwargs)
        graph = qg_builder.build(metadata=metadata, attach_datastore_records=attach_datastore_records)
        _LOG.info(
            "QuantumGraph contains %d quanta for %d tasks, graph ID: %r",
            len(graph),
            len(graph.taskGraph),
            graph.graphID,
        )
        return graph

    def build_quantum_graph(
        self,
        pipeline: Pipeline,
        where: str = "",
        *,
        builder_class: type[QuantumGraphBuilder] = AllDimensionsQuantumGraphBuilder,
        attach_datastore_records: bool = False,
        **kwargs: Any,
    ) -> PredictedQuantumGraph:
        """Build a quantum graph from a pipeline and input datasets.

        This returns an instance of the new
        `.quantum_graph.PredictedQuantumGraph` class. Use `make_quantum_graph`
        to construct a `.QuantumGraph`.

        Parameters
        ----------
        pipeline : `.Pipeline`
            The pipeline for which to generate a quantum graph.
        where : `str`, optional
            A data ID query that constrains the quanta generated.  Must not be
            provided if a custom ``builder_class`` is given and that class does
            not accept ``where`` as a construction argument.
        builder_class : `type` [ \
                `.quantum_graph_builder.QuantumGraphBuilder` ], optional
            Quantum graph builder implementation.  Ignored if ``builder`` is
            provided.
        attach_datastore_records : `bool`, optional
            Whether to attach datastore records.  These are currently used only
            by `lsst.daf.butler.QuantumBackedButler`, which is not used by
            `SeparablePipelineExecutor` for execution.
        **kwargs
            Additional keyword arguments are forwarded to ``builder_class``
            when a quantum graph builder instance is constructed.  All
            arguments accepted by the
            `~.quantum_graph_builder.QuantumGraphBuilder` base
            class are provided automatically (from explicit arguments to this
            method and executor attributes) and do not need to be included
            as keyword arguments.

        Returns
        -------
        graph : `.QuantumGraph`
            The quantum graph for ``.Pipeline`` as run on the datasets
            identified by ``where``.

        Notes
        -----
        This method does no special handling of empty quantum graphs. If
        needed, clients can use `len` to test if the returned graph is empty.
        """
        metadata = {
            "skip_existing_in": self._skip_existing_in,
            "skip_existing": bool(self._skip_existing_in),
            "data_query": where,
        }
        qg_builder = self.make_quantum_graph_builder(pipeline, where, builder_class=builder_class, **kwargs)
        graph = qg_builder.finish(
            metadata=metadata, attach_datastore_records=attach_datastore_records
        ).assemble()
        _LOG.info(
            "PredictedQuantumGraph contains %d quanta for %d tasks.",
            len(graph),
            len(graph.quanta_by_task),
        )
        return graph

    def run_pipeline(
        self,
        graph: QuantumGraph | PredictedQuantumGraph,
        fail_fast: bool = False,
        graph_executor: QuantumGraphExecutor | None = None,
        num_proc: int = 1,
        *,
        provenance_dataset_ref: DatasetRef | None = None,
    ) -> None:
        """Run a pipeline in the form of a prepared quantum graph.

        Pre-execution initialization must have already been run;
        see `pre_execute_qgraph`.

        Parameters
        ----------
        graph : `.QuantumGraph` or `.quantum_graph.PredictedQuantumGraph`
            The pipeline and datasets to execute.
        fail_fast : `bool`, optional
            If `True`, abort all execution if any task fails when
            running with multiple processes. Only used with the default graph
            executor).
        graph_executor : `.quantum_graph_executor.QuantumGraphExecutor`,\
                optional
            A custom graph executor. By default, a new instance of
            `.mp_graph_executor.MPGraphExecutor` is used.
        num_proc : `int`, optional
            The number of processes that can be used to run the pipeline. The
            default value ensures that no subprocess is created. Only used with
            the default graph executor.
        provenance_dataset_ref : `lsst.daf.butler.DatasetRef`, optional
            Dataset that should be used to save provenance.  Provenance is only
            supported when running in a single process (at least for the
            default quantum executor), and should not be used with
            ``skip_existing_in=[output_run]`` when retrying a previous
            execution attempt. The caller is responsible for registering the
            dataset type and for ensuring that the dimensions of this dataset
            do not lead to uniqueness conflicts.
        """
        if not graph_executor:
            quantum_executor = SingleQuantumExecutor(
                butler=self._butler,
                task_factory=self._task_factory,
                skip_existing_in=self._skip_existing_in,
                clobber_outputs=self._clobber_output,
                resources=self.resources,
                raise_on_partial_outputs=self.raise_on_partial_outputs,
            )
            graph_executor = MPGraphExecutor(
                num_proc=num_proc,
                timeout=2_592_000.0,  # In practice, timeout is never helpful; set to 30 days.
                quantum_executor=quantum_executor,
                fail_fast=fail_fast,
            )
            # Have to reset connection pool to avoid sharing connections with
            # forked processes.
            self._butler.registry.resetConnectionPool()

        if provenance_dataset_ref is not None:
            with TemporaryForIngest(self._butler, provenance_dataset_ref) as temporary:
                graph_executor.execute(graph, provenance_graph_file=temporary.ospath)
                temporary.ingest()
        else:
            graph_executor.execute(graph)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ["ExceptionInfo", "ExecutionStatus", "QuantumReport", "Report"]

import enum
import sys
import uuid
from typing import Any

import pydantic

from lsst.daf.butler import DataCoordinate, DataId, DataIdValue
from lsst.utils.introspection import get_full_type_name

from .graph import QgraphSummary


def _serializeDataId(dataId: DataId) -> dict[str, DataIdValue]:
    if isinstance(dataId, DataCoordinate):
        return dict(dataId.required)
    else:
        return dataId  # type: ignore


class ExecutionStatus(enum.Enum):
    """Possible values for job execution status.

    Status `FAILURE` is set if one or more tasks failed. Status `TIMEOUT` is
    set if there are no failures but one or more tasks timed out. Timeouts can
    only be detected in multi-process mode, child task is killed on timeout
    and usually should have non-zero exit code.
    """

    SUCCESS = "success"
    FAILURE = "failure"
    TIMEOUT = "timeout"
    SKIPPED = "skipped"


class ExceptionInfo(pydantic.BaseModel):
    """Information about exception."""

    className: str
    """Name of the exception class if exception was raised."""

    message: str
    """Exception message for in-process quantum execution, None if
    quantum was executed in sub-process.
    """

    @classmethod
    def from_exception(cls, exception: Exception) -> ExceptionInfo:
        """Construct instance from an exception.

        Parameters
        ----------
        exception : `Exception`
            Exception to wrap.

        Returns
        -------
        info : `ExceptionInfo`
            Information about the exception.
        """
        return cls(className=get_full_type_name(exception), message=str(exception))

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)


class QuantumReport(pydantic.BaseModel):
    """Task execution report for a single Quantum.

    Parameters
    ----------
    dataId : `~lsst.daf.butler.DataId`
        Quantum data ID.
    taskLabel : `str`
        Label for task executing this Quantum.
    status : `ExecutionStatus`
        Status of this quantum execution.
    exitCode : `int` or `None`, optional
        Exit code for sub-process executing this Quantum. `None` for
        in-process execution. Negative if process was killed by a signal.
    exceptionInfo : `ExceptionInfo` or `None`, optional
        Exception information if an exception was raised.
    quantumId : `uuid.UUID`, optional
        Unique identifier for the quantum.
    """

    status: ExecutionStatus = ExecutionStatus.SUCCESS
    """Execution status, one of the values in `ExecutionStatus` enum."""

    dataId: dict[str, DataIdValue]
    """Quantum DataId."""

    taskLabel: str | None
    """Label for a task executing this Quantum."""

    exitCode: int | None = None
    """Exit code for a sub-process executing Quantum, None for in-process
    Quantum execution. Negative if process was killed by a signal.
    """

    exceptionInfo: ExceptionInfo | None = None
    """Exception information if exception was raised."""

    quantumId: uuid.UUID | None = None
    """Unique identifier for the quantum."""

    def __init__(
        self,
        dataId: DataId,
        taskLabel: str,
        status: ExecutionStatus = ExecutionStatus.SUCCESS,
        exitCode: int | None = None,
        exceptionInfo: ExceptionInfo | None = None,
        quantumId: uuid.UUID | None = None,
    ):
        super().__init__(
            quantumId=quantumId,
            status=status,
            dataId=_serializeDataId(dataId),
            taskLabel=taskLabel,
            exitCode=exitCode,
            exceptionInfo=exceptionInfo,
        )

    @classmethod
    def from_exception(
        cls,
        exception: Exception,
        dataId: DataId,
        taskLabel: str,
        *,
        exitCode: int | None = None,
        quantumId: uuid.UUID | None = None,
    ) -> QuantumReport:
        """Construct report instance from an exception and other pieces of
        data.

        Parameters
        ----------
        exception : `Exception`
            Exception caught from processing quantum.
        dataId : `~lsst.daf.butler.DataId`
            Data ID of quantum.
        taskLabel : `str`
            Label of task.
        exitCode : `int`, optional
            Exit code for the process, used when it is known that the process
            will exit with that exit code.
        quantumId : `uuid.UUID`, optional
            Unique identifier for the quantum.
        """
        return cls(
            status=ExecutionStatus.FAILURE,
            dataId=dataId,
            taskLabel=taskLabel,
            exitCode=exitCode,
            exceptionInfo=ExceptionInfo.from_exception(exception),
            quantumId=quantumId,
        )

    @classmethod
    def from_exit_code(
        cls,
        exitCode: int,
        dataId: DataId,
        taskLabel: str,
        quantumId: uuid.UUID | None = None,
    ) -> QuantumReport:
        """Construct report instance from an exit code and other pieces of
        data.

        Parameters
        ----------
        exitCode : `int`
            The exit code of the subprocess.
        dataId : `~lsst.daf.butler.DataId`
            The quantum Data ID.
        taskLabel : `str`
            The task label.
        quantumId : `uuid.UUID`, optional
            Unique identifier for the quantum.
        """
        return cls(
            status=ExecutionStatus.SUCCESS if exitCode == 0 else ExecutionStatus.FAILURE,
            dataId=dataId,
            taskLabel=taskLabel,
            exitCode=exitCode,
            quantumId=quantumId,
        )

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)


class Report(pydantic.BaseModel):
    """Execution report for the whole job with one or few quanta."""

    qgraphSummary: QgraphSummary
    """Summary report about QuantumGraph."""

    status: ExecutionStatus = ExecutionStatus.SUCCESS
    """Job status."""

    cmdLine: list[str] | None = None
    """Command line for the whole job."""

    exitCode: int | None = None
    """Job exit code, this obviously cannot be set in pipetask."""

    exceptionInfo: ExceptionInfo | None = None
    """Exception information if exception was raised."""

    quantaReports: list[QuantumReport] = []
    """List of per-quantum reports, ordering is not specified. Some or all
    quanta may not produce a report.
    """

    # Always want to validate the default value for cmdLine so
    # use a model_validator.
    @pydantic.model_validator(mode="before")
    @classmethod
    def _set_cmdLine(cls, data: Any) -> Any:
        if data.get("cmdLine") is None:
            data["cmdLine"] = sys.argv
        return data

    def set_exception(self, exception: Exception) -> None:
        """Update exception information from an exception object.

        Parameters
        ----------
        exception : `Exception`
            Exception to use to extract information from.
        """
        self.exceptionInfo = ExceptionInfo.from_exception(exception)

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining few methods to generate Mermaid charts from pipelines or
quantum graphs.
"""

from __future__ import annotations

__all__ = ["graph2mermaid", "pipeline2mermaid"]

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal

from .pipeline import Pipeline

if TYPE_CHECKING:
    from .graph import QuantumGraph
    from .pipeline import TaskDef
    from .quantum_graph import PredictedQuantumGraph


def graph2mermaid(qgraph: QuantumGraph | PredictedQuantumGraph, file: Any) -> None:
    """Convert QuantumGraph into a Mermaid flowchart (top-down).

    This method is mostly for documentation/presentation purposes.

    Parameters
    ----------
    qgraph : `~lsst.pipe.base.QuantumGraph`
        QuantumGraph instance.
    file : `str` or file object
        File where Mermaid flowchart is written, can be a file name or file
        object.

    Raises
    ------
    OSError
        Raised if the output file cannot be opened.
    ImportError
        Raised if the task class cannot be imported.
    """
    from .quantum_graph import PredictedQuantumGraph, visualization

    if not isinstance(qgraph, PredictedQuantumGraph):
        qgraph = PredictedQuantumGraph.from_old_quantum_graph(qgraph)

    # Open a file if needed.
    close = False
    if not hasattr(file, "write"):
        file = open(file, "w")
        close = True

    v = visualization.QuantumGraphMermaidVisualizer()
    v.write_bipartite(qgraph, file)

    if close:
        file.close()


def pipeline2mermaid(
    pipeline: Pipeline | Iterable[TaskDef],
    file: Any,
    show_dimensions: bool = True,
    expand_dimensions: bool = False,
    show_storage: bool = True,
) -> None:
    """Convert a Pipeline into a Mermaid flowchart diagram.

    This function produces a Mermaid flowchart, representing tasks and their
    inputs/outputs as dataset nodes. It uses a top-down layout.

    This method is mostly for documentation/presentation purposes.

    Parameters
    ----------
    pipeline : Pipeline or Iterable[TaskDef]
        The pipeline or collection of tasks to represent.
    file : str or file-like
        The output file or file-like object into which the Mermaid code is
        written.
    show_dimensions : bool, optional
        If True, display dimension information for tasks and datasets.
        Default is True.
    expand_dimensions : bool, optional
        If True, expand dimension names to include all components. Default is
        False.
    show_storage : bool, optional
        If True, display storage class information for datasets. Default is
        True.

    Raises
    ------
    OSError
        Raised if the output file cannot be opened.
    ImportError
        Raised if the task class cannot be imported.
    """
    from .pipeline_graph import PipelineGraph, visualization

    # Ensure that pipeline is iterable of task definitions.
    if isinstance(pipeline, Pipeline):
        pipeline = pipeline.to_graph()._iter_task_defs()

    # Open file if needed.
    close = False
    if not hasattr(file, "write"):
        file = open(file, "w")
        close = True

    if isinstance(pipeline, Pipeline):
        pg = pipeline.to_graph(visualization_only=True)
    else:
        pg = PipelineGraph()
        for task_def in pipeline:
            pg.add_task(
                task_def.label,
                task_class=task_def.taskClass,
                config=task_def.config,
                connections=task_def.connections,
            )
        pg.resolve(visualization_only=True)

    dimensions: Literal["full", "concise"] | None = None
    if show_dimensions:
        if expand_dimensions:
            dimensions = "full"
        else:
            dimensions = "concise"

    visualization.show_mermaid(
        pg, stream=file, dataset_types=True, dimensions=dimensions, storage_classes=show_storage
    )

    if close:
        file.close()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["zip_from_graph"]

import logging

from lsst.daf.butler import QuantumBackedButler
from lsst.pipe.base import QuantumGraph
from lsst.resources import ResourcePath

from .utils import filter_by_dataset_type_glob

_LOG = logging.getLogger(__name__)


def zip_from_graph(
    graph: str,
    repo: str,
    dest: str,
    dataset_type: tuple[str, ...],
) -> ResourcePath:
    """Create Zip export file from graph outputs.

    Parameters
    ----------
    graph : `str`
        URI string of the quantum graph.
    repo : `str`
        URI to a butler configuration used to define the datastore associated
        with the graph.
    dest : `str`
        Path to the destination directory for the Zip file.
    dataset_type : `tuple` of `str`
        Dataset type names. An empty tuple implies all dataset types.
        Can include globs.

    Returns
    -------
    zip_path : `lsst.resources.ResourcePath`
        Path to the Zip file.
    """
    # Read whole graph into memory
    qgraph = QuantumGraph.loadUri(graph)

    output_refs, _ = qgraph.get_refs(include_outputs=True, include_init_outputs=True, conform_outputs=True)

    # Get data repository dataset type definitions from the QuantumGraph.
    dataset_types = {dstype.name: dstype for dstype in qgraph.registryDatasetTypes()}

    # Filter the refs based on requested dataset types.
    filtered_refs = filter_by_dataset_type_glob(output_refs, dataset_type)

    # Make QBB, its config is the same as output Butler.
    with QuantumBackedButler.from_predicted(
        config=repo,
        predicted_inputs=[ref.id for ref in output_refs],
        predicted_outputs=[],
        dimensions=qgraph.universe,
        datastore_records={},
        dataset_types=dataset_types,
    ) as qbb:
        _LOG.info("Retrieving artifacts for %d datasets and storing in Zip file.", len(filtered_refs))
        zip = qbb.retrieve_artifacts_zip(filtered_refs, dest)
    return zip



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["SingleQuantumExecutor"]

import logging
import time
import uuid
from collections import defaultdict
from collections.abc import Callable, Mapping
from itertools import chain
from typing import Any

from lsst.daf.butler import (
    Butler,
    DatasetRef,
    DatasetType,
    LimitedButler,
    NamedKeyDict,
    Quantum,
)
from lsst.daf.butler.logging import ButlerLogRecords
from lsst.utils.introspection import get_full_type_name
from lsst.utils.timer import logInfo

from ._quantumContext import ExecutionResources, QuantumContext
from ._status import (
    AnnotatedPartialOutputsError,
    ExceptionInfo,
    InvalidQuantumError,
    NoWorkFound,
    QuantumSuccessCaveats,
)
from .connections import AdjustQuantumHelper
from .log_capture import LogCapture, _ExecutionLogRecordsExtra
from .pipeline_graph import TaskNode
from .pipelineTask import PipelineTask
from .quantum_graph_executor import QuantumExecutionResult, QuantumExecutor
from .quantum_reports import QuantumReport
from .task import _TASK_FULL_METADATA_TYPE, _TASK_METADATA_TYPE
from .taskFactory import TaskFactory

_LOG = logging.getLogger(__name__)


class SingleQuantumExecutor(QuantumExecutor):
    """Executor class which runs one Quantum at a time.

    Parameters
    ----------
    butler : `~lsst.daf.butler.Butler` or `None`, optional
        Data butler, `None` means that a limited butler should be used instead.
    task_factory : `.TaskFactory`, optional
        Instance of a task factory.  Defaults to a new instance of
        `lsst.pipe.base.TaskFactory`.
    skip_existing_in : `str` or `~collections.abc.Iterable` [ `str` ]
        A collection name or list of collections to search for the existing
        outputs of quanta, which indicates that those quanta should be skipped.
        This class only checks for the presence of butler output run in the
        list of collections. If the output run is present in the list then the
        quanta whose complete outputs exist in the output run will be skipped.
        `None` or empty string/sequence disables skipping.
    clobber_outputs : `bool`, optional
        If `True`, then outputs from a quantum that exist in output run
        collection will be removed prior to executing a quantum. If
        ``skip_existing_in`` contains output run, then only partial outputs
        from a quantum will be removed. Only used when ``butler`` is not
        `None`.
    enable_lsst_debug : `bool`, optional
        Enable debugging with ``lsstDebug`` facility for a task.
    limited_butler_factory : `~collections.abc.Callable`, optional
        A method that creates a `~lsst.daf.butler.LimitedButler` instance for a
        given Quantum. This parameter must be defined if ``butler`` is `None`.
        If ``butler`` is not `None` then this parameter is ignored.
    resources : `.ExecutionResources`, optional
        The resources available to this quantum when executing.
    skip_existing : `bool`, optional
        If `True`, skip quanta whose metadata datasets are already stored.
        Unlike ``skip_existing_in``, this works with limited butlers as well as
        full butlers.  Always set to `True` if ``skip_existing_in`` matches
        ``butler.run``.
    assume_no_existing_outputs : `bool`, optional
        If `True`, assume preexisting outputs are impossible (e.g. because this
        is known by higher-level code to be a new ``RUN`` collection), and do
        not look for them.  This causes the ``skip_existing`` and
        ``clobber_outputs`` options to be ignored, but unlike just setting both
        of those to `False`, it also avoids all dataset existence checks.
    raise_on_partial_outputs : `bool`, optional
        If `True` raise exceptions chained by `.AnnotatedPartialOutputsError`
        immediately, instead of considering the partial result a success and
        continuing to run downstream tasks.
    job_metadata : `~collections.abc.Mapping`
        Mapping with extra metadata to embed within the quantum metadata under
        the "job" key.  This is intended to correspond to information common
        to all quanta being executed in a single process, such as the time
        taken to load the quantum graph in a BPS job.
    """

    def __init__(
        self,
        *,
        butler: Butler | None = None,
        task_factory: TaskFactory | None = None,
        skip_existing_in: Any = None,
        clobber_outputs: bool = False,
        enable_lsst_debug: bool = False,
        limited_butler_factory: Callable[[Quantum], LimitedButler] | None = None,
        resources: ExecutionResources | None = None,
        skip_existing: bool = False,
        assume_no_existing_outputs: bool = False,
        raise_on_partial_outputs: bool = True,
        job_metadata: Mapping[str, int | str | float] | None = None,
    ):
        self._butler = butler
        self._task_factory = task_factory if task_factory is not None else TaskFactory()
        self._clobber_outputs = clobber_outputs
        self._enable_lsst_debug = enable_lsst_debug
        self._limited_butler_factory = limited_butler_factory
        self._resources = resources
        self._assume_no_existing_outputs = assume_no_existing_outputs
        self._raise_on_partial_outputs = raise_on_partial_outputs
        self._job_metadata = job_metadata

        if self._butler is None:
            assert limited_butler_factory is not None, "limited_butler_factory is needed when butler is None"

        # Find whether output run is in skip_existing_in.
        self._skip_existing = skip_existing
        if self._butler is not None and skip_existing_in and not self._skip_existing:
            self._skip_existing = self._butler.run in self._butler.collections.query(
                skip_existing_in, flatten_chains=True
            )
        self._previous_process_quanta: list[uuid.UUID] = []

    def execute(
        self,
        task_node: TaskNode,
        /,
        quantum: Quantum,
        quantum_id: uuid.UUID | None = None,
        *,
        log_records: ButlerLogRecords | None = None,
    ) -> QuantumExecutionResult:
        # Docstring inherited from QuantumExecutor.execute
        if self._butler is not None:
            self._butler.registry.refresh()
        return self._execute(task_node, quantum, quantum_id=quantum_id, log_records=log_records)

    def _execute(
        self,
        task_node: TaskNode,
        /,
        quantum: Quantum,
        quantum_id: uuid.UUID | None = None,
        *,
        log_records: ButlerLogRecords | None = None,
    ) -> QuantumExecutionResult:
        """Execute the quantum.

        Internal implementation of `execute()`.
        """
        # Make a limited butler instance if needed.
        limited_butler: LimitedButler
        used_butler_factory = False
        if self._butler is not None:
            limited_butler = self._butler
        else:
            # We check this in constructor, but mypy needs this check here.
            assert self._limited_butler_factory is not None
            limited_butler = self._limited_butler_factory(quantum)
            used_butler_factory = True

        try:
            return self._execute_with_limited_butler(
                task_node, limited_butler, quantum=quantum, quantum_id=quantum_id, log_records=log_records
            )
        finally:
            if used_butler_factory:
                limited_butler.close()

    def _execute_with_limited_butler(
        self,
        task_node: TaskNode,
        limited_butler: LimitedButler,
        /,
        quantum: Quantum,
        quantum_id: uuid.UUID | None = None,
        *,
        log_records: ButlerLogRecords | None = None,
    ) -> QuantumExecutionResult:
        startTime = time.time()
        assert quantum.dataId is not None, "Quantum DataId cannot be None"
        report = QuantumReport(quantumId=quantum_id, dataId=quantum.dataId, taskLabel=task_node.label)
        if self._butler is not None:
            log_capture = LogCapture.from_full(self._butler)
        else:
            log_capture = LogCapture.from_limited(limited_butler)
        with log_capture.capture_logging(task_node, quantum, records=log_records) as captureLog:
            # Save detailed resource usage before task start to metadata.
            quantumMetadata = _TASK_METADATA_TYPE()
            logInfo(None, "prep", metadata=quantumMetadata)  # type: ignore[arg-type]

            _LOG.info(
                "Preparing execution of quantum for label=%s dataId=%s.", task_node.label, quantum.dataId
            )

            # check whether to skip or delete old outputs, if it returns True
            # or raises an exception do not try to store logs, as they may be
            # already in butler.
            captureLog.store = False
            if self._check_existing_outputs(quantum, task_node, limited_butler, captureLog.extra):
                _LOG.info(
                    "Skipping already-successful quantum for label=%s dataId=%s.",
                    task_node.label,
                    quantum.dataId,
                )
                return QuantumExecutionResult(quantum, report, skipped_existing=True, adjusted_no_work=False)
            captureLog.store = True

            captureLog.extra.previous_process_quanta.extend(self._previous_process_quanta)
            if quantum_id is not None:
                self._previous_process_quanta.append(quantum_id)
            try:
                quantum = self._updated_quantum_inputs(quantum, task_node, limited_butler)
            except NoWorkFound as exc:
                _LOG.info(
                    "Nothing to do for task '%s' on quantum %s; saving metadata and skipping: %s",
                    task_node.label,
                    quantum.dataId,
                    str(exc),
                )
                quantumMetadata["caveats"] = QuantumSuccessCaveats.from_adjust_quantum_no_work().value
                quantumMetadata["outputs"] = []
                # Make empty metadata that looks something like what a
                # do-nothing task would write (but we don't bother with empty
                # nested PropertySets for subtasks).  This is slightly
                # duplicative with logic in pipe_base that we can't easily call
                # from here; we'll fix this on DM-29761.
                logInfo(None, "end", metadata=quantumMetadata)  # type: ignore[arg-type]
                fullMetadata = _TASK_FULL_METADATA_TYPE()
                fullMetadata[task_node.label] = _TASK_METADATA_TYPE()
                fullMetadata["quantum"] = quantumMetadata
                if self._job_metadata is not None:
                    fullMetadata["job"] = self._job_metadata
                self._write_metadata(quantum, fullMetadata, task_node, limited_butler)
                return QuantumExecutionResult(
                    quantum,
                    report,
                    skipped_existing=False,
                    adjusted_no_work=True,
                    task_metadata=fullMetadata,
                )

            # enable lsstDebug debugging
            if self._enable_lsst_debug:
                try:
                    _LOG.debug("Will try to import debug.py")
                    import debug  # type: ignore # noqa:F401
                except ImportError:
                    _LOG.warning("No 'debug' module found.")

            # Ensure that we are executing a frozen config
            task_node.config.freeze()
            logInfo(None, "init", metadata=quantumMetadata)  # type: ignore[arg-type]
            init_input_refs = list(quantum.initInputs.values())

            _LOG.info(
                "Constructing task and executing quantum for label=%s dataId=%s.",
                task_node.label,
                quantum.dataId,
            )
            task = self._task_factory.makeTask(task_node, limited_butler, init_input_refs)
            logInfo(None, "start", metadata=quantumMetadata)  # type: ignore[arg-type]
            outputs_put: list[uuid.UUID] = []
            try:
                with limited_butler.record_metrics() as butler_metrics:
                    caveats = self._run_quantum(
                        task, quantum, task_node, limited_butler, quantum_id=quantum_id, ids_put=outputs_put
                    )
            except Exception as e:
                _LOG.error(
                    "Execution of task '%s' on quantum %s failed. Exception %s: %s",
                    task_node.label,
                    quantum.dataId,
                    e.__class__.__name__,
                    str(e),
                )
                captureLog.extra.exception = ExceptionInfo(
                    type_name=get_full_type_name(e),
                    message=str(e),
                    metadata={},
                )
                raise
            else:
                quantumMetadata["butler_metrics"] = butler_metrics.model_dump()
                quantumMetadata["caveats"] = caveats.value
                # Stringify the UUID for easier compatibility with
                # PropertyList.
            finally:
                logInfo(None, "end", metadata=quantumMetadata)  # type: ignore[arg-type]
                fullMetadata = task.getFullMetadata()
                quantumMetadata["outputs"] = [str(output) for output in outputs_put]
                fullMetadata["quantum"] = quantumMetadata
                if self._job_metadata is not None:
                    fullMetadata["job"] = self._job_metadata
                captureLog.extra.metadata = fullMetadata
            self._write_metadata(quantum, fullMetadata, task_node, limited_butler)
            stopTime = time.time()
            _LOG.info(
                "Execution of task '%s' on quantum %s took %.3f seconds",
                task_node.label,
                quantum.dataId,
                stopTime - startTime,
            )
        return QuantumExecutionResult(
            quantum,
            report,
            skipped_existing=False,
            adjusted_no_work=False,
            task_metadata=fullMetadata,
        )

    def _check_existing_outputs(
        self,
        quantum: Quantum,
        task_node: TaskNode,
        /,
        limited_butler: LimitedButler,
        log_extra: _ExecutionLogRecordsExtra,
    ) -> bool:
        """Decide whether this quantum needs to be executed.

        If only partial outputs exist then they are removed if
        ``clobberOutputs`` is True, otherwise an exception is raised.

        The ``LimitedButler`` is used for everything, and should be set to
        ``self.butler`` if no separate ``LimitedButler`` is available.

        Parameters
        ----------
        quantum : `~lsst.daf.butler.Quantum`
            Quantum to check for existing outputs.
        task_node : `~.pipeline_graph.TaskNode`
            Task definition structure.
        limited_butler : `~lsst.daf.butler.LimitedButler`
            Butler to use for querying and clobbering.
        log_extra : `.log_capture.TaskLogRecordsExtra`
            Extra information to attach to log records.

        Returns
        -------
        exist : `bool`
            `True` if ``self.skipExisting`` is defined, and a previous
            execution of this quanta appears to have completed successfully
            (either because metadata was written or all datasets were written).
            `False` otherwise.

        Raises
        ------
        RuntimeError
            Raised if some outputs exist and some not.
        """
        if self._assume_no_existing_outputs:
            return False

        if self._skip_existing:
            _LOG.debug(
                "Checking existence of metadata from previous execution of label=%s dataId=%s.",
                task_node.label,
                quantum.dataId,
            )
            # Metadata output exists; this is sufficient to assume the previous
            # run was successful and should be skipped.
            [metadata_ref] = quantum.outputs[task_node.metadata_output.dataset_type_name]
            if metadata_ref is not None:
                if limited_butler.stored(metadata_ref):
                    return True

        # Find and prune (partial) outputs if `self.clobberOutputs` is set.
        _LOG.debug(
            "Looking for existing outputs in the way for label=%s dataId=%s.", task_node.label, quantum.dataId
        )
        ref_dict = limited_butler.stored_many(chain.from_iterable(quantum.outputs.values()))
        if task_node.log_output is not None:
            (log_ref,) = quantum.outputs[task_node.log_output.dataset_type_name]
            if ref_dict[log_ref]:
                _LOG.debug(
                    "Attaching logs from previous attempt on label=%s dataId=%s.",
                    task_node.label,
                    quantum.dataId,
                )
                log_extra.attach_previous_attempt(limited_butler.get(log_ref))
        existingRefs = [ref for ref, exists in ref_dict.items() if exists]
        missingRefs = [ref for ref, exists in ref_dict.items() if not exists]
        if existingRefs:
            if not missingRefs:
                # Full outputs exist.
                if self._skip_existing:
                    return True
                elif self._clobber_outputs:
                    _LOG.info("Removing complete outputs for quantum %s: %s", quantum, existingRefs)
                    limited_butler.pruneDatasets(existingRefs, disassociate=True, unstore=True, purge=True)
                else:
                    raise RuntimeError(
                        f"Complete outputs exists for a quantum {quantum} "
                        "and neither clobberOutputs nor skipExisting is set: "
                        f"existingRefs={existingRefs}"
                    )
            else:
                # Partial outputs from a failed quantum.
                _LOG.debug(
                    "Partial outputs exist for quantum %s existingRefs=%s missingRefs=%s",
                    quantum,
                    existingRefs,
                    missingRefs,
                )
                if self._clobber_outputs:
                    # only prune
                    _LOG.info("Removing partial outputs for task %s: %s", task_node.label, existingRefs)
                    limited_butler.pruneDatasets(existingRefs, disassociate=True, unstore=True, purge=True)
                    return False
                else:
                    raise RuntimeError(
                        "Registry inconsistency while checking for existing quantum outputs:"
                        f" quantum={quantum} existingRefs={existingRefs}"
                        f" missingRefs={missingRefs}"
                    )

        # By default always execute.
        return False

    def _updated_quantum_inputs(
        self, quantum: Quantum, task_node: TaskNode, /, limited_butler: LimitedButler
    ) -> Quantum:
        """Update quantum with extra information, returns a new updated
        Quantum.

        Some methods may require input DatasetRefs to have non-None
        ``dataset_id``, but in case of intermediate dataset it may not be
        filled during QuantumGraph construction. This method will retrieve
        missing info from registry.

        Parameters
        ----------
        quantum : `~lsst.daf.butler.Quantum`
            Single Quantum instance.
        task_node : `~.pipeline_graph.TaskNode`
            Task definition structure.
        limited_butler : `~lsst.daf.butler.LimitedButler`
            Butler to use for querying.

        Returns
        -------
        update : `~lsst.daf.butler.Quantum`
            Updated Quantum instance.
        """
        anyChanges = False
        updatedInputs: defaultdict[DatasetType, list] = defaultdict(list)
        for key, refsForDatasetType in quantum.inputs.items():
            _LOG.debug(
                "Checking existence of input '%s' for label=%s dataId=%s.",
                key.name,
                task_node.label,
                quantum.dataId,
            )
            toCheck = []
            newRefsForDatasetType = updatedInputs[key]
            for ref in refsForDatasetType:
                if self._should_assume_exists(quantum, ref):
                    newRefsForDatasetType.append(ref)
                else:
                    toCheck.append(ref)
            if not toCheck:
                _LOG.debug(
                    "Assuming overall input '%s' is present without checks for label=%s dataId=%s.",
                    key.name,
                    task_node.label,
                    quantum.dataId,
                )
                continue
            stored = limited_butler.stored_many(toCheck)
            for ref in toCheck:
                if stored[ref]:
                    newRefsForDatasetType.append(ref)
                else:
                    # This should only happen if a predicted intermediate was
                    # not actually produced upstream, but
                    # datastore misconfigurations can unfortunately also land
                    # us here.
                    _LOG.info("No dataset artifact found for %s", ref)
                    continue
            if len(newRefsForDatasetType) != len(refsForDatasetType):
                anyChanges = True
        # If we removed any input datasets, let the task check if it has enough
        # to proceed and/or prune related datasets that it also doesn't
        # need/produce anymore.  It will raise NoWorkFound if it can't run,
        # which we'll let propagate up.  This is exactly what we run during QG
        # generation, because a task shouldn't care whether an input is missing
        # because some previous task didn't produce it, or because it just
        # wasn't there during QG generation.
        namedUpdatedInputs = NamedKeyDict[DatasetType, list[DatasetRef]](updatedInputs.items())
        helper = AdjustQuantumHelper(namedUpdatedInputs, quantum.outputs)
        if anyChanges:
            _LOG.debug("Running adjustQuantum for label=%s dataId=%s.", task_node.label, quantum.dataId)
            assert quantum.dataId is not None, "Quantum DataId cannot be None"
            helper.adjust_in_place(task_node.get_connections(), label=task_node.label, data_id=quantum.dataId)
        return Quantum(
            taskName=quantum.taskName,
            taskClass=quantum.taskClass,
            dataId=quantum.dataId,
            initInputs=quantum.initInputs,
            inputs=helper.inputs,
            outputs=helper.outputs,
        )

    def _run_quantum(
        self,
        task: PipelineTask,
        quantum: Quantum,
        task_node: TaskNode,
        /,
        limited_butler: LimitedButler,
        quantum_id: uuid.UUID | None,
        ids_put: list[uuid.UUID],
    ) -> QuantumSuccessCaveats:
        """Execute task on a single quantum.

        Parameters
        ----------
        task : `PipelineTask`
            Task object.
        quantum : `~lsst.daf.butler.Quantum`
            Single Quantum instance.
        task_node : `~.pipeline_graph.TaskNode`
            Task definition structure.
        limited_butler : `~lsst.daf.butler.LimitedButler`
            Butler to use for dataset I/O.
        quantum_id : `uuid.UUID` or `None`
            ID of the quantum being executed.
        ids_put : list[ `uuid.UUID` ]
            List to be populated with the dataset IDs that were written by this
            quantum.  This is an output parameter in order to allow it to be
            populated even when an exception is raised.

        Returns
        -------
        flags : `QuantumSuccessCaveats`
            Flags that describe qualified successes.
        """
        flags = QuantumSuccessCaveats.NO_CAVEATS

        # Create a butler that operates in the context of a quantum
        butlerQC = QuantumContext(limited_butler, quantum, resources=self._resources, quantum_id=quantum_id)

        # Get the input and output references for the task
        inputRefs, outputRefs = task_node.get_connections().buildDatasetRefs(quantum)

        # Call task runQuantum() method.
        try:
            task.runQuantum(butlerQC, inputRefs, outputRefs)
        except NoWorkFound as err:
            # Not an error, just an early exit.
            _LOG.info(
                "Task '%s' on quantum %s exited early with no work found: %s.",
                task_node.label,
                quantum.dataId,
                str(err),
            )
            flags |= err.FLAGS
        except AnnotatedPartialOutputsError as caught:
            error: BaseException
            if caught.__cause__ is None:
                _LOG.error(
                    "Incorrect use of AnnotatedPartialOutputsError: no chained exception found.",
                    task_node.label,
                    quantum.dataId,
                )
                error = caught
            else:
                error = caught.__cause__
            if self._raise_on_partial_outputs:
                # Note: this is a real edge case that required some
                # experimentation: without 'from None' below, this raise would
                # produce a "while one exception was being handled, another was
                # raised" traceback involving AnnotatedPartialOutputsError.
                # With the 'from None', we get just the error chained to it, as
                # desired.
                raise error from None
            else:
                _LOG.error(
                    "Task '%s' on quantum %s exited with partial outputs; "
                    "considering this a qualified success and proceeding.",
                    task_node.label,
                    quantum.dataId,
                )
                _LOG.error(error, exc_info=error)
                flags |= caught.FLAGS
        finally:
            ids_put.extend(output[2] for output in butlerQC.outputsPut)
        if not butlerQC.outputsPut:
            flags |= QuantumSuccessCaveats.ALL_OUTPUTS_MISSING
        if not butlerQC.outputsPut == butlerQC.allOutputs:
            flags |= QuantumSuccessCaveats.ANY_OUTPUTS_MISSING
        return flags

    def _write_metadata(
        self, quantum: Quantum, metadata: Any, task_node: TaskNode, /, limited_butler: LimitedButler
    ) -> None:
        # DatasetRef has to be in the Quantum outputs, can lookup by name
        try:
            [ref] = quantum.outputs[task_node.metadata_output.dataset_type_name]
        except LookupError as exc:
            raise InvalidQuantumError(
                "Quantum outputs is missing metadata dataset type "
                f"{task_node.metadata_output.dataset_type_name};"
                " this could happen due to inconsistent options between QuantumGraph generation"
                " and execution"
            ) from exc
        limited_butler.put(metadata, ref)

    def _should_assume_exists(self, quantum: Quantum, ref: DatasetRef) -> bool | None:
        """Report whether the given dataset can be assumed to exist because
        some previous check reported that it did.

        If this is `True` for a dataset does not in fact exist anymore, that's
        an unexpected problem that we want to raise as an exception, and
        definitely not a case where some predicted output just wasn't produced.
        We can't always tell the difference, but in this case we can.

        Parameters
        ----------
        quantum : `Quantum`
            Quantum being processed.
        ref : `lsst.daf.butler.DatasetRef`
            Reference to the input dataset.

        Returns
        -------
        exists : `bool` or `None`
            `True` if this dataset is definitely an overall input, `False` if
            some other quantum in the graph is expected to produce it, and
            `None` if the answer could not be determined.
        """
        if quantum.datastore_records:
            for datastore_record_data in quantum.datastore_records.values():
                if ref.id in datastore_record_data.records:
                    return True
            return False
        return None



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("AggregatorConfig",)

import sys
from typing import TYPE_CHECKING, Any

import pydantic


class AggregatorConfig(pydantic.BaseModel):
    """Configuration for the provenance aggregator."""

    output_path: str | None = None
    """Path for the output provenance quantum graph file.

    At present this option is intended only for debugging.
    """

    worker_log_dir: str | None = None
    """Path to a directory (POSIX only) for parallel worker logs."""

    worker_log_level: str = "VERBOSE"
    """Log level for worker processes/threads.

    Per-quantum messages only appear at ``DEBUG`` level.
    """

    worker_profile_dir: str | None = None
    """Path to a directory (POSIX only) for parallel worker profiling dumps.

    This option is ignored when `n_processes` is `1`.
    """

    n_processes: int = 1
    """Number of processes the scanner should use."""

    incomplete: bool = False
    """If `True`, do not expect the graph to have been executed to completion
    yet, and only ingest the outputs of successful quanta.

    This disables writing the provenance quantum graph, since this is likely to
    be wasted effort that just complicates a follow-up run with
    ``incomplete=False`` later.
    """

    defensive_ingest: bool = False
    """If `True`, guard against datasets having already been ingested into the
    central butler repository.

    Defensive ingest mode is automatically turned on (with a warning emitted)
    if an ingest attempt fails due to a database constraint violation. Enabling
    defensive mode up-front avoids this warning and is slightly more efficient
    when it is already known that some datasets have already been ingested.

    Defensive mode does not guard against race conditions from multiple ingest
    processes running simultaneously, as it relies on a one-time query to
    determine what is already present in the central repository.
    """

    ingest_batch_size: int = 10000
    """Number of butler datasets that must accumulate to trigger an ingest."""

    register_dataset_types: bool = True
    """Whether to register output dataset types in the central butler
    repository before starting ingest.
    """

    update_output_chain: bool = True
    """Whether to prepend the output `~lsst.daf.butler.CollectionType.RUN` to
    the output `~lsst.daf.butler.CollectionType.CHAINED` collection.
    """

    dry_run: bool = False
    """If `True`, do not actually perform any central butler ingests.

    Most log messages concerning ingests will still be emitted in order to
    provide a better emulation of a real run.
    """

    interactive_status: bool = False
    """Whether to use an interactive status display with progress bars.

    If this is `True`, the `tqdm` module must be available.  If this is
    `False`, a periodic logger will be used to display status at a fixed
    interval instead (see `log_status_interval`).
    """

    log_status_interval: float | None = None
    """Interval (in seconds) between periodic logger status updates."""

    worker_sleep: float = 0.01
    """Time (in seconds) a worker should wait when there are no requests from
    the main aggregator process.
    """

    zstd_level: int = 10
    """ZStandard compression level to use for all compressed-JSON blocks."""

    zstd_dict_size: int = 32768
    """Size (in bytes) of the ZStandard compression dictionary."""

    zstd_dict_n_inputs: int = 512
    """Number of samples of each type (see below) to include in ZStandard
    compression dictionary training.

    Training is run on a random subset of the `PredictedQuantumDatasetsModel`
    objects in the predicted graph, as well as the first provenance quanta,
    logs, and metadata blocks encountered.
    """

    mock_storage_classes: bool = False
    """Enable support for storage classes by created by the
    lsst.pipe.base.tests.mocks package.
    """

    promise_ingest_graph: bool = False
    """If `True`, the aggregator will assume that `~.ingest_graph.ingest_graph`
    will be run later to ingest metadata/log/config datasets, and will not
    ingest them itself.  This means that if `~.ingest_graph.ingest_graph` is
    not run, those files will be abandoned in the butler storage root without
    being present in the butler database, but it will speed up both processes.

    It is *usually* safe to build a quantum graph for downstream processing
    before or while running `~.ingest_graph.ingest_graph`, because
    metadata/log/config datasets are rarely used as inputs.  To check, use
    ``pipetask build ... --show inputs`` to show the overall-inputs to the
    graph and scan for these dataset types.
    """

    worker_check_timeout: float = 5.0
    """Time to wait (s) for reports from subprocesses before running
    process-alive checks.

    These checks are designed to kill the main aggregator process when a
    subprocess has been unexpectedly killed (e.g. for for using too much
    memory).
    """

    @property
    def is_writing_provenance(self) -> bool:
        """Whether the aggregator is configured to write the provenance quantum
        graph.
        """
        return self.output_path is not None and not self.incomplete

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules and not TYPE_CHECKING:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("SimplePipelineExecutor",)

import os
from collections.abc import Iterable, Iterator, Mapping
from typing import Any

from lsst.daf.butler import (
    Butler,
    CollectionType,
    DataCoordinate,
    DatasetRef,
    Quantum,
)
from lsst.daf.butler.registry import RegistryDefaults
from lsst.pex.config import Config

from ._instrument import Instrument
from ._quantumContext import ExecutionResources
from .all_dimensions_quantum_graph_builder import AllDimensionsQuantumGraphBuilder
from .graph import QuantumGraph
from .pipeline import Pipeline
from .pipeline_graph import PipelineGraph
from .pipelineTask import PipelineTask
from .quantum_graph import PredictedQuantumGraph
from .single_quantum_executor import SingleQuantumExecutor
from .taskFactory import TaskFactory


class SimplePipelineExecutor:
    """A simple, high-level executor for pipelines.

    Parameters
    ----------
    quantum_graph : `.QuantumGraph`
        Graph to be executed.
    butler : `~lsst.daf.butler.Butler`
        Object that manages all I/O.  Must be initialized with `collections`
        and `run` properties that correspond to the input and output
        collections, which must be consistent with those used to create
        ``quantum_graph``.
    resources : `.ExecutionResources`
        The resources available to each quantum being executed.
    raise_on_partial_outputs : `bool`, optional
        If `True` raise exceptions chained by `.AnnotatedPartialOutputsError`
        immediately, instead of considering the partial result a success and
        continuing to run downstream tasks.

    Notes
    -----
    Most callers should use one of the `classmethod` factory functions
    (`from_pipeline_filename`, `from_task_class`, `from_pipeline`) instead of
    invoking the constructor directly; these guarantee that the
    `~lsst.daf.butler.Butler` and `.QuantumGraph` are created consistently.

    This class is intended primarily to support unit testing and small-scale
    integration testing of `.PipelineTask` classes.  It deliberately lacks many
    features present in the command-line-only ``pipetask`` tool in order to
    keep the implementation simple.  Python callers that need more
    sophistication should call lower-level tools like
    `~.quantum_graph_builder.QuantumGraphBuilder` and
    `.single_quantum_executor.SingleQuantumExecutor` directly.
    """

    def __init__(
        self,
        quantum_graph: QuantumGraph | PredictedQuantumGraph,
        butler: Butler,
        resources: ExecutionResources | None = None,
        raise_on_partial_outputs: bool = True,
    ):
        from .graph import QuantumGraph

        self._quantum_graph: QuantumGraph | None = None
        if isinstance(quantum_graph, QuantumGraph):
            self._quantum_graph = quantum_graph
            self.predicted = PredictedQuantumGraph.from_old_quantum_graph(self._quantum_graph)
        else:
            self.predicted = quantum_graph
        self.butler = butler
        self.resources = resources
        self.raise_on_partial_outputs = raise_on_partial_outputs

    @classmethod
    def prep_butler(
        cls,
        root: str,
        inputs: Iterable[str],
        output: str,
        output_run: str | None = None,
    ) -> Butler:
        """Return configured `~lsst.daf.butler.Butler`.

        Helper method for creating `~lsst.daf.butler.Butler` instances with
        collections appropriate for processing.

        Parameters
        ----------
        root : `str`
            Root of the butler data repository; must already exist, with all
            necessary input data.
        inputs : `~collections.abc.Iterable` [ `str` ]
            Collections to search for all input datasets, in search order.
        output : `str`
            Name of a new output `~lsst.daf.butler.CollectionType.CHAINED`
            collection to create that will combine both inputs and outputs.
        output_run : `str`, optional
            Name of the output `~lsst.daf.butler.CollectionType.RUN` that will
            directly hold all output datasets.  If not provided, a name will be
            created from ``output`` and a timestamp.

        Returns
        -------
        butler : `~lsst.daf.butler.Butler`
            Butler client instance compatible with all `classmethod` factories.
            Always writeable.
        """
        if output_run is None:
            output_run = f"{output}/{Instrument.makeCollectionTimestamp()}"
        # Make initial butler with no collections, since we haven't created
        # them yet.
        butler = Butler.from_config(root, writeable=True)
        butler.registry.registerCollection(output_run, CollectionType.RUN)
        butler.registry.registerCollection(output, CollectionType.CHAINED)
        collections = [output_run]
        collections.extend(inputs)
        butler.registry.setCollectionChain(output, collections)
        # Override the registry defaults. No need to clone.
        butler.registry.defaults = RegistryDefaults(collections=[output], run=output_run)
        return butler

    @classmethod
    def from_pipeline_filename(
        cls,
        pipeline_filename: str,
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
        resources: ExecutionResources | None = None,
        raise_on_partial_outputs: bool = True,
        attach_datastore_records: bool = False,
        output: str | None = None,
        output_run: str | None = None,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from an on-disk
        pipeline YAML file.

        Parameters
        ----------
        pipeline_filename : `str`
            Name of the YAML file to load the pipeline definition from.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `~collections.abc.Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.
        resources : `.ExecutionResources`
            The resources available to each quantum being executed.
        raise_on_partial_outputs : `bool`, optional
            If `True` raise exceptions chained by
            `.AnnotatedPartialOutputsError` immediately, instead of considering
            the partial result a success and continuing to run downstream
            tasks.
        attach_datastore_records : `bool`, optional
            Whether to attach datastore records to the quantum graph.  This is
            usually unnecessary, unless the executor is used to test behavior
            that depends on datastore records.
        output : `str`, optional
            Name of a new output `~lsst.daf.butler.CollectionType.CHAINED`
            collection to create that will combine both inputs and outputs.
        output_run : `str`, optional
            Name of the output `~lsst.daf.butler.CollectionType.RUN` that will
            directly hold all output datasets.  If not provided, a name will be
            created from ``output`` and a timestamp.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed `.QuantumGraph` and
            `~lsst.daf.butler.Butler`, ready for `run` to be called.
        """
        pipeline = Pipeline.fromFile(pipeline_filename)
        return cls.from_pipeline(
            pipeline,
            butler=butler,
            where=where,
            bind=bind,
            resources=resources,
            raise_on_partial_outputs=raise_on_partial_outputs,
            attach_datastore_records=attach_datastore_records,
            output=output,
            output_run=output_run,
        )

    @classmethod
    def from_task_class(
        cls,
        task_class: type[PipelineTask],
        config: Config | None = None,
        label: str | None = None,
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
        resources: ExecutionResources | None = None,
        raise_on_partial_outputs: bool = True,
        attach_datastore_records: bool = False,
        output: str | None = None,
        output_run: str | None = None,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from a pipeline
        containing a single task.

        Parameters
        ----------
        task_class : `type`
            A concrete `.PipelineTask` subclass.
        config : `~lsst.pex.config.Config`, optional
            Configuration for the task.  If not provided, task-level defaults
            will be used (no per-instrument overrides).
        label : `str`, optional
            Label for the task in its pipeline; defaults to
            ``task_class._DefaultName``.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `~collections.abc.Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.
        resources : `.ExecutionResources`
            The resources available to each quantum being executed.
        raise_on_partial_outputs : `bool`, optional
            If `True` raise exceptions chained by
            `.AnnotatedPartialOutputsError` immediately, instead of considering
            the partial result a success and continuing to run downstream
            tasks.
        attach_datastore_records : `bool`, optional
            Whether to attach datastore records to the quantum graph.  This is
            usually unnecessary, unless the executor is used to test behavior
            that depends on datastore records.
        output : `str`, optional
            Name of a new output `~lsst.daf.butler.CollectionType.CHAINED`
            collection to create that will combine both inputs and outputs.
        output_run : `str`, optional
            Name of the output `~lsst.daf.butler.CollectionType.RUN` that will
            directly hold all output datasets.  If not provided, a name will be
            created from ``output`` and a timestamp.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed `.QuantumGraph` and
            `~lsst.daf.butler.Butler`, ready for `run` to be called.
        """
        if config is None:
            config = task_class.ConfigClass()
        if label is None:
            label = task_class._DefaultName
        if not isinstance(config, task_class.ConfigClass):
            raise TypeError(
                f"Invalid config class type: expected {task_class.ConfigClass.__name__}, "
                f"got {type(config).__name__}."
            )
        pipeline_graph = PipelineGraph()
        pipeline_graph.add_task(label=label, task_class=task_class, config=config)
        return cls.from_pipeline_graph(
            pipeline_graph,
            butler=butler,
            where=where,
            bind=bind,
            resources=resources,
            raise_on_partial_outputs=raise_on_partial_outputs,
            attach_datastore_records=attach_datastore_records,
            output=output,
            output_run=output_run,
        )

    @classmethod
    def from_pipeline(
        cls,
        pipeline: Pipeline,
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
        resources: ExecutionResources | None = None,
        raise_on_partial_outputs: bool = True,
        attach_datastore_records: bool = False,
        output: str | None = None,
        output_run: str | None = None,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from an in-memory
        pipeline.

        Parameters
        ----------
        pipeline : `.Pipeline` or `~collections.abc.Iterable` [ `.TaskDef` ]
            A Python object describing the tasks to run, along with their
            labels and configuration.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `~collections.abc.Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.
        resources : `.ExecutionResources`
            The resources available to each quantum being executed.
        raise_on_partial_outputs : `bool`, optional
            If `True` raise exceptions chained by
            `.AnnotatedPartialOutputsError` immediately, instead of considering
            the partial result a success and continuing to run downstream
            tasks.
        attach_datastore_records : `bool`, optional
            Whether to attach datastore records to the quantum graph.  This is
            usually unnecessary, unless the executor is used to test behavior
            that depends on datastore records.
        output : `str`, optional
            Name of a new output `~lsst.daf.butler.CollectionType.CHAINED`
            collection to create that will combine both inputs and outputs.
        output_run : `str`, optional
            Name of the output `~lsst.daf.butler.CollectionType.RUN` that will
            directly hold all output datasets.  If not provided, a name will
            be created from ``output`` and a timestamp.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed `.QuantumGraph` and
            `~lsst.daf.butler.Butler`, ready for `run` to be called.
        """
        pipeline_graph = pipeline.to_graph()
        return cls.from_pipeline_graph(
            pipeline_graph,
            where=where,
            bind=bind,
            butler=butler,
            resources=resources,
            raise_on_partial_outputs=raise_on_partial_outputs,
            attach_datastore_records=attach_datastore_records,
            output=output,
            output_run=output_run,
        )

    @classmethod
    def from_pipeline_graph(
        cls,
        pipeline_graph: PipelineGraph,
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
        resources: ExecutionResources | None = None,
        raise_on_partial_outputs: bool = True,
        attach_datastore_records: bool = False,
        output: str | None = None,
        output_run: str | None = None,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from an in-memory
        pipeline graph.

        Parameters
        ----------
        pipeline_graph : `~.pipeline_graph.PipelineGraph`
            A Python object describing the tasks to run, along with their
            labels and configuration, in graph form.  Will be resolved against
            the given ``butler``, with any existing resolutions ignored.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `~collections.abc.Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.  Must have its `~lsst.daf.butler.Butler.run` and
            ``butler.collections.defaults`` not empty and not `None`.
        resources : `.ExecutionResources`
            The resources available to each quantum being executed.
        raise_on_partial_outputs : `bool`, optional
            If `True` raise exceptions chained by
            `.AnnotatedPartialOutputsError` immediately, instead
            of considering the partial result a success and continuing to run
            downstream tasks.
        attach_datastore_records : `bool`, optional
            Whether to attach datastore records to the quantum graph.  This is
            usually unnecessary, unless the executor is used to test behavior
            that depends on datastore records.
        output : `str`, optional
            Name of a new output `~lsst.daf.butler.CollectionType.CHAINED`
            collection to create that will combine both inputs and outputs.
        output_run : `str`, optional
            Name of the output `~lsst.daf.butler.CollectionType.RUN` that will
            directly hold all output datasets.  If not provided, a name will
            be created from ``output`` and a timestamp.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed
            `.QuantumGraph` and `~lsst.daf.butler.Butler`, ready
            for `run` to be called.
        """
        if output_run is None:
            output_run = butler.run
            if output_run is None:
                if output is None:
                    raise TypeError("At least one of output or output_run must be provided.")
                output_run = f"{output}/{Instrument.makeCollectionTimestamp()}"

        quantum_graph_builder = AllDimensionsQuantumGraphBuilder(
            pipeline_graph, butler, where=where, bind=bind, output_run=output_run
        )
        metadata = {
            "skip_existing_in": [],
            "skip_existing": False,
            "data_query": where,
        }
        predicted = quantum_graph_builder.finish(
            output=output,
            metadata=metadata,
            attach_datastore_records=attach_datastore_records,
        ).assemble()
        return cls(
            predicted,
            butler=butler,
            resources=resources,
            raise_on_partial_outputs=raise_on_partial_outputs,
        )

    @property
    def quantum_graph(self) -> QuantumGraph:
        """The quantum graph run by this executor."""
        if self._quantum_graph is None:
            self._quantum_graph = self.predicted.to_old_quantum_graph()
        return self._quantum_graph

    def use_local_butler(
        self, root: str, register_dataset_types: bool = True, transfer_dimensions: bool = True
    ) -> Butler:
        """Transfer all inputs to a local data repository. and set the executor
        to write outputs to it.

        Parameters
        ----------
        root : `str`
            Path to the local data repository; created if it does not exist.
        register_dataset_types : `bool`, optional
            Whether to register dataset types in the new repository.  If
            `False`, the local data repository must already exist and already
            have all input dataset types registered.
        transfer_dimensions : `bool`, optional
            Whether to transfer dimension records to the new repository.  If
            `False`, the local data repository must already exist and already
            have all needed dimension records.

        Returns
        -------
        butler : `lsst.daf.butler.Butler`
            Writeable butler for local data repository.

        Notes
        -----
        The input collection structure from the original data repository is not
        preserved by this method (it cannot be reconstructed from the quantum
        graph).  Instead, a `~lsst.daf.butler.CollectionType.TAGGED` collection
        is created to gather all inputs, and appended to the output
        `~lsst.daf.butler.CollectionType.CHAINED` collection after the output
        `~lsst.daf.butler.CollectionType.RUN` collection.  Calibration inputs
        with the same data ID but multiple validity ranges are *not* included
        in that `~lsst.daf.butler.CollectionType.TAGGED`; they are still
        transferred to the local data repository, but can only be found via the
        quantum graph or their original `~lsst.daf.butler.CollectionType.RUN`
        collections.
        """
        if not os.path.exists(root):
            Butler.makeRepo(root)
        out_butler = Butler.from_config(root, writeable=True)

        output_run = self.predicted.header.output_run
        out_butler.collections.register(output_run, CollectionType.RUN)
        output = self.predicted.header.output
        inputs: str | None = None
        if output is not None:
            inputs = f"{output}/inputs"
            out_butler.collections.register(output, CollectionType.CHAINED)
            out_butler.collections.register(inputs, CollectionType.TAGGED)
            out_butler.collections.redefine_chain(output, [output_run, inputs])

        if transfer_dimensions:
            # We can't just let the transfer_from call below take care of this
            # because we need dimensions for outputs as well as inputs.  And if
            # we have to do the outputs explicitly, it's more efficient to do
            # the inputs at the same time since a lot of those dimensions will
            # be the same.
            self._transfer_qg_dimension_records(out_butler)

        # Extract overall-input DatasetRefs to transfer and possibly insert
        # into a TAGGED collection.
        refs: set[DatasetRef] = set()
        to_tag_by_type: dict[str, dict[DataCoordinate, DatasetRef | None]] = {}
        pipeline_graph = self.predicted.pipeline_graph
        for name, dataset_type_node in pipeline_graph.iter_overall_inputs():
            assert dataset_type_node is not None, "PipelineGraph should be resolved."
            to_tag_for_type = to_tag_by_type.setdefault(name, {})
            for task_node in pipeline_graph.consumers_of(name):
                for quantum in self.predicted.build_execution_quanta(task_label=task_node.label).values():
                    for ref in quantum.inputs[name]:
                        ref = dataset_type_node.generalize_ref(ref)
                        refs.add(ref)
                        if to_tag_for_type.setdefault(ref.dataId, ref) != ref:
                            # There is already a dataset with the same data ID
                            # and dataset type, but a different UUID/run.  This
                            # can only happen for calibrations found in
                            # calibration collections, and for now we have no
                            # choice but to leave them out of the TAGGED inputs
                            # collection in the local butler.
                            to_tag_for_type[ref.dataId] = None

        out_butler.transfer_from(
            self.butler,
            refs,
            register_dataset_types=register_dataset_types,
            transfer_dimensions=False,
        )

        if inputs is not None:
            to_tag_flat: list[DatasetRef] = []
            for ref_map in to_tag_by_type.values():
                for tag_ref in ref_map.values():
                    if tag_ref is not None:
                        to_tag_flat.append(tag_ref)
            out_butler.registry.associate(inputs, to_tag_flat)

        out_butler.registry.defaults = self.butler.registry.defaults.clone(collections=output, run=output_run)
        self.butler = out_butler
        return self.butler

    def run(self, register_dataset_types: bool = False, save_versions: bool = True) -> list[Quantum]:
        """Run all the quanta in the quantum graph in topological order.

        Use this method to run all quanta in the graph.  Use
        `as_generator` to get a generator to run the quanta one at
        a time.

        Parameters
        ----------
        register_dataset_types : `bool`, optional
            If `True`, register all output dataset types before executing any
            quanta.
        save_versions : `bool`, optional
            If `True` (default), save a package versions dataset.

        Returns
        -------
        quanta : `list` [ `~lsst.daf.butler.Quantum` ]
            Executed quanta.

        Notes
        -----
        A topological ordering is not in general unique, but no other
        guarantees are made about the order in which quanta are processed.
        """
        return list(
            self.as_generator(register_dataset_types=register_dataset_types, save_versions=save_versions)
        )

    def as_generator(
        self, register_dataset_types: bool = False, save_versions: bool = True
    ) -> Iterator[Quantum]:
        """Yield quanta in the quantum graph in topological order.

        These quanta will be run as the returned generator is iterated
        over.  Use this method to run the quanta one at a time.
        Use `run` to run all quanta in the graph.

        Parameters
        ----------
        register_dataset_types : `bool`, optional
            If `True`, register all output dataset types before executing any
            quanta.
        save_versions : `bool`, optional
            If `True` (default), save a package versions dataset.

        Returns
        -------
        quanta : `~collections.abc.Iterator` [ `~lsst.daf.butler.Quantum` ]
            Executed quanta.

        Notes
        -----
        Global initialization steps (see `.QuantumGraph.init_output_run`) are
        performed immediately when this method is called, but individual quanta
        are not actually executed until the returned iterator is iterated over.

        A topological ordering is not in general unique, but no other
        guarantees are made about the order in which quanta are processed.
        """
        if register_dataset_types:
            self.predicted.pipeline_graph.register_dataset_types(self.butler)
        self.predicted.write_configs(self.butler, compare_existing=False)
        self.predicted.write_init_outputs(self.butler, skip_existing=False)
        if save_versions:
            self.predicted.write_packages(self.butler, compare_existing=False)
        task_factory = TaskFactory()
        single_quantum_executor = SingleQuantumExecutor(
            butler=self.butler,
            task_factory=task_factory,
            resources=self.resources,
            raise_on_partial_outputs=self.raise_on_partial_outputs,
        )
        self.predicted.build_execution_quanta()
        nodes_map = self.predicted.quantum_only_xgraph.nodes
        # Important that this returns a generator expression rather than being
        # a generator itself; that is what makes the init stuff above happen
        # immediately instead of when the first quanta is executed, which might
        # be useful for callers who want to check the state of the repo in
        # between.
        return (
            single_quantum_executor.execute(
                nodes_map[quantum_id]["pipeline_node"],
                nodes_map[quantum_id]["quantum"],
                quantum_id,
            )[0]
            for quantum_id in self.predicted
        )

    def _transfer_qg_dimension_records(self, out_butler: Butler) -> None:
        """Transfer all dimension records from the quantum graph to a butler.

        Parameters
        ----------
        out_butler : `lsst.daf.butler.Butler`
            Butler to transfer records to.
        """
        assert self.predicted.dimension_data is not None, "Dimension data must be present for execution."
        records = self.predicted.dimension_data.records
        dimensions = out_butler.dimensions.sorted(records.keys())
        for dimension in dimensions:
            record_set = records[dimension.name]
            if record_set and record_set.element.has_own_table:
                out_butler.registry.insertDimensionData(
                    record_set.element,
                    *record_set,
                    skip_existing=True,
                )



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A tool for aggregating provenance information, logs, and metadata from a
processing run into a `ProvenanceQuantumGraph`.

The aggregator uses multiple processes (typically) or threads to gather
information about a processing run that has already completed, ingest the
output datasets into a butler repository, and write provenance information to
disk.

The only public entry point is the `aggregate_graph` function, with all options
controlled by an `AggregatorConfig` instance.
Usually the tool is invoked from the command line as
``butler aggregate-graph``.

Most log messages are sent to the ``aggregate-graph`` log, which users may want
to set to ``VERBOSE`` to get additional detail, including periodic logging.
See `AggregatorConfig` for how to enable additional logging by each process.
"""

from __future__ import annotations

__all__ = ("AggregatorConfig", "FatalWorkerError", "aggregate_graph")

from ._config import AggregatorConfig
from ._supervisor import aggregate_graph
from ._communicators import FatalWorkerError

#
# Aggregator Design Notes and Package Layout
# ==========================================
#
# The aggregator tool is designed to run in multiple processes.  Using threads
# instead is supported, but mostly as a way to leverage the GIL to (mostly)
# only occupy a single core without having to maintain two largely distinct
# implementations.  Throughout the rest of these notes I'll just assume they're
# processes.
#
# There are four types of processes, each represented in code as a class in its
# own module:
#
# - The Supervisor runs in the main process and is responsible for interacting
#   with the user (reporting progress, handling KeyboardInterrupt) and
#   traversing the predicted quantum graph.
#
# - Scanner workers are responsible for reading log and metadata files using a
#   QBB, occasionally checking for the existence of other outputs (we can
#   usually get the information about which outputs exist from the metadata)
#   and sending the information we extract from them to other workers. Scanners
#   are the only kind of worker we can add more of when we are given more
#   cores, so we try to do as much pre-processing as possible on them (e.g.
#   compression, datastore record extraction).
#
# - The ingester worker is responsible for ingesting output datasets into the
#   central butler repository.  It batches up ingest requests from the scanners
#   until the batch size exceeds a configurable limit (e.g. 10k datasets), and
#   then ingests those pending requests in a single transaction.
#
# - The writer worker is responsible for actually writing the
#   `ProvenanceQuantumGraph` file.  Unlike the other workers, it needs to read
#   almost the entire predicted quantum graph file before it starts, so it can
#   take a few minutes before it starts its main loop of accepting write
#   requests from the scanners (most of the time) and the supervisor (only for
#   blocked quanta, which we don't bother to scan).  For each of the big
#   multi-block files (datasets, quanta, logs, metadata) the writes opens and
#   writes to a temporary file, which is moved into the provenance QG zip
#   archive at the end.  In addition, to try to keep expensive operations on
#   the scanners, the writer builds a ZStandard compression dictionary from the
#   first write requests it sees, and then sends that to the scanners so they
#   can take over doing the compression after that is complete. Since writing
#   the provenance graph is (at present) optional, sometimes there is no writer
#   worker.
#
# Each process is also associated with its own "communicator" class, which are
# all defined together in the _communicators.py file.  The communicators manage
# the queues that are used to send information between processes, as well as
# providing loggers and some other common utility functionality. Communicators
# are context managers that take care of cleanly shutting down each process
# when it completes or is interrupted; `multiprocessing.Queue` objects need to
# be cleared out carefully or you get deadlocks that prevent processes from
# actually exiting, and so a lot of the work of the communicators involves
# passing sentinal objects over those queues so we can identify when they
# really are empty.
#
# The other modules in the packages include:
#
# - _config.py: the `AggregatorConfig` model, which is essentially the public
#   interface for controlling the aggregator.
#
# - _progress.py: helper objects for [periodic] logging and tqdm progress bars.
#   It's not clear the tqdm stuff should continue to exist once development of
#   the tool is complete and we stop wanting to run it from an interactive
#   terminal, but it's certainly been useful for development so far.
#
# - _structs.py: simple dataclasses used by multiple workers (most of what we
#   put on the queues to request scans/ingests/writes);
#
# The main logger used by the aggregator is called
# ``lsst.pipe.base.quantum_graph.aggregator``, and it emits at ``VERBOSE`` and
# above.  At ``VERBOSE``, it's also used for periodic reporting of status
# (unless the tqdm progress bars are enabled instead). Each worker also has its
# own log (including the supervisor) with names like
# ``lsst.pipe.base.quantum_graph.aggregator.scanner-012`` that include
# additional information at ``DEBUG`` and above, but these are *not* propagated
# up; instead there's a config option to write per-worker logs to separate
# files in a directory.  It would probably be more idiomatic to set up a
# QueueHandler to feed all the LogRecords the supervisor and use Python logging
# configuration to send them wherever desired, but::
#
# - that seemed to require more wading into `lsst.daf.butler.cli.cliLog` a lot
#   more than I cared to (especially considering that we want this to work in
#   `threading` and `multiprocessing`, and logging configuration is very
#   different in those two contexts);
#
# - having the worker logs go to separate files is actually very nice, and it's
#   more efficient if they just do that themselves, and that's not something
#   our logging CLI can actually do, AFAICT.
